; Vi-like editor and brainfuck interpreter in 512 bytes
; It was a nice way to waste time

org 0x7c00
use16


; Registers
start:
  sti

  mov ax, 0x1000
  mov ds, ax
  mov es, ax
  shl ax, 1
  mov fs, ax
  mov sp, 0x7c00
  xor di, di
  mov ss, di
  mov ah, 0xb8
  mov gs, ax

file_clear:
  cld
  
  mov ax, 0x2020
  mov cx, 0x8000
  rep stosw

  xor dx, dx

; Vi like command mode
; accepts 'o' - opens a new line
;         'i' - enters insert mode
;         'h' - moves the cursor left
;         'l' - moves the cursor right
command_mode:
  mov bp, 0x743
  call draw_screen
.loop:
  call move_cursor
  xor ax, ax
  int 0x16
  cmp al, 'i'
  je input_mode
  cmp al, 'h'
  jne .l
  dec dl
  jns .l
  inc dl
.l:
  cmp al, 'l'
  jne .I
  cmp dl, 79
  je .loop
  inc dl
.I:
  cmp al, 'I'
  jne .o
  call interpret
  jmp short command_mode
.o:
  cmp al, 'o'
  jne .O
  add di, 80
  cmp dh, 23
  je .o1
  inc dh
.o1:
  jmp short .Oo
.O:
  cmp al, 'O'
  jne .j
.Oo:
  call add_line
  jmp short input_mode
.j:
  cmp al, 'j'
  jne .k
  add di, 80
  cmp dh, 23
  je .update_screen
  inc dh
.k:
  cmp al, 'k'
  jne .loop
  cmp di, 0
  je .loop
  sub di, 80
  cmp dh, 0 
  je .update_screen
  dec dh
.update_screen:
  jmp short command_mode


; Input mode
; Esc brings you out of it
input_mode:
  mov bp, 0x749
  call draw_screen
.loop:
  call move_cursor
  xor ax, ax
  int 0x16
  cmp al, 0x1b
  je command_mode

  cmp al, 8
  jne .no_backspace
  cmp dl, 0
  je .loop
  dec dl
  
  call line_del_char
  jmp short input_mode

.no_backspace:
.line_add_char:
  cmp dl, 80
  je .loop

  mov cx, 79
  sub cl, dl
  mov si, di
  add si, 79

  cmp al, 13
  jne .no_newline
  sub si, cx
  add di, 80
  push di
  push cx
  push si
  cmp dh, 23
  je .no_inc
  inc dh
.no_inc:
  call add_line
  pop si
  pop cx

  push cx
  push si
  rep movsb
  pop di
  pop cx
  mov al, ' '
  rep stosb
  pop di
  jmp short input_mode

.no_newline:
  push di
  mov di, si
  dec si
  std
  rep movsb
  cld
  mov [di], al
  pop di
  inc dl

.redraw:
  jmp short input_mode


; Deletes a character from the line (di) at dl
line_del_char:
  push di
  mov cx, 79
  movzx si, dl
  sub cx, si
  add di, si
  mov si, di
  inc si
  rep movsb
  mov byte[di], ' '
  pop di
  ret


; insert a new line before the current one (kept in di)
; and sets di to the new line
add_line:
  push di
  xor cx, cx
  sub cx, di
  dec cx
  mov di, -1
  mov si, -81
  std
  rep movsb
  cld

  mov cl, 80
  mov al, ' '
  pop di
  rep stosb
  sub di, 80
  
  xor dl, dl
  call draw_screen
  ret


; Moves the cursor to (x, y) = (dl, dh)
move_cursor:
  mov ah, 2
  xor bx, bx
  int 0x10
  ret


; Draws all visible lines
draw_screen:
  call clear_screen
  push di
  movzx ax, dh
  cmp ax, 0
  je .loop1_end
.loop1:
  sub di, 80
  dec ax
  jnz .loop1
.loop1_end:
  
  mov cx, 24*80
  mov ah, 7
  xor bx, bx
.loop2:
  mov al, [di]
  mov [gs:bx], ax
  inc di
  inc bx
  inc bx
  
  loop .loop2

  mov bx, 24*80*2
  mov [gs:bx], bp
  
  pop di
  ret


; Clears the screen (no kidding!)
clear_screen:
  mov si, 80*25*2-2
.loop:
  mov word [gs:si], 0x720
  dec si
  dec si
  jns .loop
  ret

; Teh brainfuck interpreter
interpret:
  push di
  push dx

  xor di, di
  xor bx, bx
  xor dx, dx

  call move_cursor
  call clear_screen

.loop2:
  mov byte[fs:bx], 0
  inc bx
  jnz .loop2

  mov cx, 0xffff
.loop3:
  mov al, byte[di]
  cmp al, '['
  jne .1
  cmp dx, 0
  je .01
  inc dx 
  jmp short .8
.01:  
  cmp byte[fs:bx], 0
  jne .02
  inc dx 
  jmp short .8
.02:
  push si
  push di
  push cx
.1:

  cmp al, ']'
  jne .2
  cmp dx, 0
  je .12
  dec dx 
  jmp short .8
.12:
  pop cx
  pop di
  pop si
  jmp short .loop3

.2:

  cmp dx, 0
  jne .8

  cmp al, '<'
  jne .3
  dec bx

.3:
  cmp al, '>'
  jne .4
  inc bx

.4:
  cmp al, '+'
  jne .5
  inc byte[fs:bx]

.5:
  cmp al, '-'
  jne .6
  dec byte[fs:bx]

.6:
  cmp al, ','
  jne .7
  xor ax, ax
  int 0x16
  cmp al, 13
  jne .no_newline
  mov al, 10
.no_newline:
  mov [fs:bx], al
  jmp short .print_char
  
.7:
  cmp al, '.'
  jne .8
  mov al, [fs:bx]
.print_char:
  push bx
  mov ah, 0xe
  mov bx, 7
  int 0x10
  pop bx

.8:
  
  inc di
  loop .loop3

  ; wait for key press
  xor ax, ax
  int 0x16
  
  pop dx
  pop di
  ret

; Padding and boot signature
times 510 - ($ - start) db 0
dw 0xaa55
